﻿using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;

namespace MVRPlugin {

	public class VisemeAligner2 : MVRScript {


        protected JSONStorableBool debugMessagesOn;
        protected JSONStorableFloat jawDriveMultiplierWhenNotSpeaking;
        protected JSONStorableFloat visemeOffsetFromAudio;

        protected AudioSourceControl headAudioSource; //The head audio source of the VAM character the plugin is attached to

        protected Hashtable visemeToFloat; // Conversions from VAM visemes names to the VAM float objects
        protected Hashtable clipDisplayNameToGentleAudio;  //List of all the audio that this person can speak

        protected JSONStorableFloat jawAudioMultiplier;     //User tweakable value for the multiplier used on the non-speech jaw audio sounds
        protected float jawDefaultMultiplierValue = 0.0f;   //The original user setting of the jaw drive value which is reset after the plug-in is done

        protected bool speechIsReady = false;       //True when the gentle files have been parsed and are ready for use

        //Conversion of ARPABet 2 character phonetic codes (left column - key) to the nearest VAM viseme morphs (right column - value)
        //VAM only has a small subset of face morphs for the relevant sounds, so they are mapped onto the nearest available morph
        //These are just my guess at how to do the conversion
        //The comment is an IPA pronunciation code for the phoneme followed by an English example where the sound being spoken is between the ~ characters
        //https://en.wikipedia.org/wiki/ARPABET

        Hashtable ARPABet2ChToVAM = new Hashtable()
        {
            {"AA","AA"}, //	ɑ	b~o~t
            {"AE","EH"}, //	æ	b~a~t
            {"AH","EH"}, //	ʌ	b~u~tt
            {"AO","UW"}, //	ɔ	b~ough~t
            {"AW","AA"}, //	aʊ	b~ou~t
            {"AX","EH"}, //	ə	~a~bout
            {"AXR","ER"}, //ɚ	lett~er~
            //{"AY","IY"}, //	aɪ 	b~i~te 
            {"AY","IH"}, //	aɪ 	b~i~te 
            {"B","M"},   //	b	~b~uy
            {"CH","SH"}, //	tʃ	~ch~ina
            {"D","T"},   //	d	~d~og
            {"DH","TH"}, //	ð	~th~y
            {"DX","T"},  // ɾ	bu~tt~er
            {"EH","EH"}, //	ɛ	b~e~t
            {"EL","T"}, //	l̩	bott~l~e
            {"EM","M"}, //	m̩	rhyth~m~
            {"EN","T"}, //	n̩	butt~on~
            {"ER","ER"}, //	ɝ	b~ir~d
            {"EY","IH"}, //	eɪ	b~ai~t
            {"F","F"}, //	f	~f~ight
            {"G","T"}, //	g	~g~uy
            {"H","K"}, //	h	~h~igh
            {"HH","K"}, //	h	~h~igh
            {"IH","IH"}, //	ɪ	b~i~t
            {"IX","IH"}, //	ɨ	ros~e~s
            {"IY","IY"}, //	i	b~ea~t
            {"JH","S"}, //	dʒ	~j~ive
            {"K","K"}, //	k	~k~ite
            {"L","L"}, //	l	~l~ie
            {"M","M"}, //	m	~m~y
            {"N","TH"}, //	n	~n~igh
            {"NG","T"}, //	ŋ	si~ng~
            {"NX","T"}, //	ŋ	si~ng~
            {"OW","OW"}, //	oʊ	b~oa~t
            {"OY","UW"}, //	ɔɪ	b~oy~
            {"P","M"}, //	p	~p~ie
            {"Q",""}, //	ʔ	glottal stop (silence)
            {"R","ER"}, //	ɹ	~r~ye
            {"S","S"}, //	s	~s~igh
            {"SH","SH"}, //	ʃ	~sh~y
            {"T","T"}, //	t	~t~ie
            {"TH","TH"}, //	θ	~th~igh
            {"UH","UW"}, //	ʊ	b~oo~k
            {"UW","UW"}, //	u	b~oo~t
            {"V","F"}, //	v	~v~ie
            {"W","W"}, //	w	~w~ise
            {"WH","W"}, //	ʍ	~wh~y
            {"Y","IY"}, //	j	~y~acht
            {"Z","T"}, //	z	~z~oo
            {"ZH","SH"}, //	ʒ	plea~s~ure

            {"","" }    //Null conversion, convert nothing to nothing
        };

        //Converts ARPABet 1 char codes into 2 char codes

        Hashtable ARPABet1ChTo2Ch = new Hashtable()
        {
            {"a","AA"},
            {"@","AE"},
            {"A","AH"},
            {"c","AO"},
            {"W","AW"},
            {"x","AX"},
            {"Y","AY"},
            {"b","B"},
            {"C","CH"},
            {"d","D"},
            {"D","DH"},
            {"F","DX"},
            {"E","EH"},
            {"L","EL"},
            {"M","EM"},
            {"N","EN"},
            {"R","ER"},
            {"e","EY"},
            {"f","F"},
            {"g","G"},
            {"h","HH"},
            {"I","IH"},
            {"X","IX"},
            {"i","IY"},
            {"J","JH"},
            {"k","K"},
            {"l","L"},
            {"m","M"},
            {"n","N"},
            {"o","OW"},
            {"O","OY"},
            {"p","P"},
            {"Q","Q"},
            {"r","R"},
            {"s","S"},
            {"S","SH"},
            {"t","T"},
            {"T","TH"},
            {"U","UH"},
            {"u","UW"},
            {"v","V"},
            {"w","W"},
            {"H","WH"},
            {"y","Y"},
            {"z","Z"},
            {"Z","ZH"}
        };

        //Gets references to the animatable viseme morphs on the person

        void setUpMorphs(Atom containingPersonAtom)
        {
            this.visemeToFloat = new Hashtable();
            List<String> receiverList = containingPersonAtom.GetStorableIDs();

            if (receiverList != null)
            {
                JSONStorable receiver = containingPersonAtom.GetStorableByID("geometry");
					
                if (receiver != null)
                {
                    DAZCharacterSelector character = receiver as DAZCharacterSelector;
                    GenerateDAZMorphsControlUI morphControl = character.morphsControlUI;
			        morphControl.GetMorphByDisplayName("AA").animatable = true;
		            morphControl.GetMorphByDisplayName("EH").animatable = true;
					morphControl.GetMorphByDisplayName("ER").animatable = true;
					morphControl.GetMorphByDisplayName("F").animatable = true;
			        morphControl.GetMorphByDisplayName("IH").animatable = true;
		            morphControl.GetMorphByDisplayName("IY").animatable = true;
					morphControl.GetMorphByDisplayName("K").animatable = true;
					morphControl.GetMorphByDisplayName("L").animatable = true;
			        morphControl.GetMorphByDisplayName("M").animatable = true;
		            morphControl.GetMorphByDisplayName("OW").animatable = true;
					morphControl.GetMorphByDisplayName("S").animatable = true;
					morphControl.GetMorphByDisplayName("SH").animatable = true;
			        morphControl.GetMorphByDisplayName("T").animatable = true;
		            morphControl.GetMorphByDisplayName("TH").animatable = true;
					morphControl.GetMorphByDisplayName("UW").animatable = true;
					morphControl.GetMorphByDisplayName("W").animatable = true;
						
                    foreach (string visemeName in receiver.GetFloatParamNames())
                    {
                        visemeToFloat.Add(visemeName, receiver.GetFloatJSONParam(visemeName));
                    }

                }
                else
                {
                    SuperController.LogError("VisemeAligner: Failed to find geometry.");
                }

           }
           else
           {
                SuperController.LogMessage("VisemeAligner: NULL RECEIVER!");
           }
		   
           this.jawAudioMultiplier = containingPersonAtom.GetStorableByID("JawControl").GetFloatJSONParam("driveXRotationFromAudioSourceMultiplier");

        //User failed to set all of the morphs to be animatable on the person	
        //All animatables now set automatically by the script so this shuoldn't happen now
        if (visemeToFloat.Count < 16)
            SuperController.LogMessage("VisemeAligner: Can't find the Visemes. Are all of the Visemes set to Animatable for this Person? You need to check all of them.");
	}
        //Holder class to map a timing to a viseme
        public class Vis2FloatSeqObj
        {
            public int riseStartMS;
            public int maximumMS;
            public int endTimeMS = 0;
            public int sustainEnd = 0;  //For singing, hold the viseme on long notes
            public string viseme_name;
            public JSONStorableFloat viseme;
        }

        //Holder class that contains audio together with the viseme timings
        public class GentleAudio
        {
            public string audioManagerDisplayName;
            public string audioManagerDisplayNameNoExtension;
            public JSONNode gentleJSON = null;
            public string gentleFileName;
            public List<Vis2FloatSeqObj> visemeSequence = null;
            public NamedAudioClip namedAudioClip;
        }

        List<Vis2FloatSeqObj> currentVis2FloatSeq;
        Hashtable activelySpoken;

        bool amSpeaking = false;
        int previousViseme = -1;
        int nextViseme = 0;
        int currentViseme = 0;
        int maxViseme = 0;
        int speakStartTimeMS = 0;
        float aggregateTimeS = 0.0f;



        public List<Vis2FloatSeqObj> encodeGentleVisemeTimings(JSONNode gentleTimings)
        {
            List<Vis2FloatSeqObj> seq = new List<Vis2FloatSeqObj>();

            try
            {
                JSONNode words = gentleTimings["words"];
                
                int lastPhonemeMaxTimeMS = -1;
                foreach (JSONNode vt in words.Childs)
                {
                    
                    JSONNode phones = vt["phones"];
                    float wordStartTimeS = vt["start"].AsFloat;
                    float phonemeMaximum = wordStartTimeS;
                    foreach (JSONNode pn in phones.Childs)
                    {
                        String gentlePhoneme = pn["phone"];

                        String[] phonemeParts = gentlePhoneme.Split('_');
                        if (phonemeParts.Length == 2)
                        {
                            //I think a gentle phoneme of OOV means "out of vocabulary" and means the speech recogniser failed
                            //So just move onto the next phoneme
                            if (!phonemeParts[0].Equals("oov"))
                            {

                                String Ch2Phoneme = phonemeParts[0].ToUpper();
                                String vamViseme = (String)ARPABet2ChToVAM[Ch2Phoneme];

                                if (vamViseme != null)
                                {

                                    Vis2FloatSeqObj v2f = new Vis2FloatSeqObj();

                                    if (visemeToFloat.ContainsKey(vamViseme))
                                    {
                                        int phonemeMaximumMS = ((int)(phonemeMaximum * 1000.0f));

                                        JSONStorableFloat jsf = (JSONStorableFloat)visemeToFloat[vamViseme];
                                        v2f.viseme = jsf;
                                        v2f.viseme_name = vamViseme;


                                        //Set the time at which the phoneme is fully on
                                        //This is the time at which Gentle recorded the phoneme start
                                        v2f.maximumMS = phonemeMaximumMS;

                                        //Set the time at which the phoneme starts rising towards being fully on
                                        if(lastPhonemeMaxTimeMS > 0)
                                        {
                                            //If this is not the first phoneme, set the start time to the point at
                                            //which the previous phoneme reached a maximum
                                            v2f.riseStartMS = lastPhonemeMaxTimeMS;

                                            //Set a maximum limit on the rise time of 330ms
                                            //This is to deal with long gaps bewteen words
                                            if((phonemeMaximumMS - v2f.riseStartMS) > 330)
                                            {
                                                v2f.riseStartMS = phonemeMaximumMS - 330;
                                                if (v2f.riseStartMS < 0)
                                                    v2f.riseStartMS = 0;
                                            }
                                        }
                                        else
                                        {
                                            //If this is the first phoneme, set the start time 100ms before the maximum
                                            v2f.riseStartMS = phonemeMaximumMS - 100;
                                            if (v2f.riseStartMS < 0)
                                                v2f.riseStartMS = 0;
                                        }



                                        //The time at which the phoneme is fully off (start time + the duration of the phoneme that gentle provided)
                                        v2f.endTimeMS = ((int)(pn["duration"].AsFloat * 1000.0f))+ v2f.maximumMS;

                                        //Sustain period if the phoneme is held for a long time (more than 250ms)
                                        //Hold the phoneme at maximum until near the end of the duration
                                        //Otherwise it will start gradually falling off through the sustain
                                        if((v2f.endTimeMS - v2f.maximumMS) > 250)
                                        {
                                            //Set the fall off to be the last 30% of the viseme
                                            int sustainPeriodEnd = (int)(  ((float)(v2f.endTimeMS - v2f.maximumMS)) * 0.70f ) + v2f.maximumMS;
                                            v2f.sustainEnd = sustainPeriodEnd;
                                        }

                                        seq.Add(v2f);

                                        lastPhonemeMaxTimeMS = v2f.maximumMS;
                                    }

                                }
                                else
                                {
                                    SuperController.LogMessage("No conversion for gentle phoneme:" + Ch2Phoneme);
                                }

                            }
                        }

                        phonemeMaximum += pn["duration"].AsFloat;
                    }

                }
            }
            catch (Exception e)
            {
                SuperController.LogError("VisemeAligner:Problem parsing the gentle file:"+e.ToString());
            }

            return seq;

        }




        public override void Init()
        {
            this.activelySpoken = new Hashtable();

            this.pluginLabelJSON.val = "VisemeAligner V2.0";	
            //UI for the debug messages check box
            this.debugMessagesOn = new JSONStorableBool("DebugMessagesOn",false);
            this.debugMessagesOn.storeType = JSONStorableParam.StoreType.Full;
            RegisterBool(this.debugMessagesOn);
            CreateToggle(this.debugMessagesOn, true);

            //Generate a trigger target on this person atom for every item of audio that has a gentle file
            getAllAudioClipsSansExtension();
            foreach (GentleAudio audio in this.clipDisplayNameToGentleAudio.Values)
            {
                RegisterAction(new JSONStorableAction("Speak:" + audio.audioManagerDisplayName,
                    new JSONStorableAction.ActionCallback(createSpeakingDelegate(audio.audioManagerDisplayName))));
            }

            RegisterAction(new JSONStorableAction("StopSpeaking", new JSONStorableAction.ActionCallback( silenceCurrentlySpeakingVisemes ) ));

            //UI for the Jaw drive multiplier slider
           this.jawDriveMultiplierWhenNotSpeaking = new JSONStorableFloat("JawDriveMultiplierWhenNotSpeaking", 50.0f, 0f, 1000.0f, true);
            this.jawDriveMultiplierWhenNotSpeaking.storeType = JSONStorableParam.StoreType.Full;
            RegisterFloat(this.jawDriveMultiplierWhenNotSpeaking);
            CreateSlider(this.jawDriveMultiplierWhenNotSpeaking, true);


        }


        //Returns a listing of all the Audio files listed in the AudioManager without the final extension
        //e.g. "CoolAudio.mp3" returns "CoolAudio"
        public void getAllAudioClipsSansExtension()
        {
            this.clipDisplayNameToGentleAudio = new Hashtable();

            string loadDir = SuperController.singleton.currentLoadDir;
			if (URLAudioClipManager.singleton.GetCategoryClips("web") != null)
			{
				foreach (NamedAudioClip clip in URLAudioClipManager.singleton.GetCategoryClips("web"))
					{
					//Strip the extension off the audio files 
					//There is a C# library method that does this - Path.GetFileNameWithoutExtension - but it's in System.io which is disabled

					GentleAudio ga = new GentleAudio();
					ga.namedAudioClip = clip;
					ga.audioManagerDisplayName = clip.displayName;

					if (clip.displayName.Contains("."))
					{
						string clipNoExt = clip.displayName.Substring(0, clip.displayName.LastIndexOf('.'));
						if (clipNoExt.Length > 0)
						{
							ga.audioManagerDisplayNameNoExtension = clipNoExt;
						}
					}
					else
						ga.audioManagerDisplayNameNoExtension = ga.audioManagerDisplayName;

					try
					{
						string jsonPath = loadDir + @"\" + ga.audioManagerDisplayNameNoExtension + ".gentle";
						JSONNode gentleTimings = LoadJSON(jsonPath);
						ga.gentleJSON = gentleTimings;
						if (this.debugMessagesOn.val)
							SuperController.LogMessage("VisemeAligner: Found gentle file found for \"" + ga.audioManagerDisplayName +"\"");

						this.clipDisplayNameToGentleAudio.Add(ga.audioManagerDisplayName, ga);
					}
					catch (Exception)
					{
						if (this.debugMessagesOn.val)
							SuperController.LogMessage("VisemeAligner: No gentle file found for \"" + ga.audioManagerDisplayName + "\".\nPlace gentle file named \"" + ga.audioManagerDisplayNameNoExtension + ".gentle\" in " + loadDir);
					}
				}				
			}
			else
			{
				SuperController.LogMessage("VisemeAligner: No audio clips loaded.");
			}
        }

        /**
         *  We should now have a list of parsed gentle JSON files.
         *  This converts the to an internal representation which is easier the code to use
         */

        public IEnumerator convertGentleJSONToAlignmentObjects()
        {

            string loadDir = SuperController.singleton.currentLoadDir;
            foreach(GentleAudio clip in this.clipDisplayNameToGentleAudio.Values) {
                //If have no method of testing if a file exists because it's disabled.
                //So the way I get round this is by trying to load it and then watching for an exception
                try
                {
                    //Generate a list of morphs and timings
                    if (clip.gentleJSON == null)
                    {
                        if (this.debugMessagesOn.val)
                            SuperController.LogMessage("VisemeAligner: ERROR: Looks like there was a problem with decoding the gentle JSON for \"" + clip.audioManagerDisplayName + "\"");
                    }
                    else
                    {
                        clip.visemeSequence = encodeGentleVisemeTimings(clip.gentleJSON);
                        //Dereference the gentle JSON parse which is no longer necessary to hold in RAM. The garbage collector should be able to sweep it now.
                        //At this stage we have parsed it and converted it to an internal representation of all the necessary information
                        clip.gentleJSON = null;

                        if(clip.visemeSequence == null)
                            if (this.debugMessagesOn.val)
                                SuperController.LogMessage("VisemeAligner: ERROR: Looks like there was a problem interpretting the gentle JSON for \"" + clip.audioManagerDisplayName + "\".\nThe JSON is probably valid but is not in the expected format.");
                    }
                }
                catch (Exception e) //Can't get the system.io exception as it's blocked, so check for any exception
                {
                    if (this.debugMessagesOn.val)
                        SuperController.LogMessage("VisemeAligner: ERROR : Unexpected exception in convertGentleJSONToAlignmentObjects \"" + clip.audioManagerDisplayName +"\""+e);
                }
                yield return null; //Avoid blocking the main thread for too long while parsing gentle files
            }

            //Tell the system that the speech is now prepared and available
            this.speechIsReady = true;
        }

        //Plays the audio and starts the visemes
        public IEnumerator startSpeaking(string audioDisplayName)
        {
            //Wait until the gentle file as actually been decoded before allowing speech to be possible
            yield return new WaitUntil(() => this.speechIsReady == true);

            if (this.clipDisplayNameToGentleAudio.ContainsKey(audioDisplayName))
            {
                GentleAudio audio = (GentleAudio)this.clipDisplayNameToGentleAudio[audioDisplayName];

                silenceCurrentlySpeakingVisemes();

                List<Vis2FloatSeqObj> encodedPollyVisemeTimings = audio.visemeSequence;
                NamedAudioClip nac = audio.namedAudioClip;

                this.jawDefaultMultiplierValue = this.jawAudioMultiplier.val;   //preserve the original jaw drive setting
                this.jawAudioMultiplier.val = 0.0f;
                this.currentVis2FloatSeq = encodedPollyVisemeTimings;
                this.maxViseme = encodedPollyVisemeTimings.Count - 1;
                this.previousViseme = -1;
                this.nextViseme = 0;
                this.currentViseme = 0;
                this.speakStartTimeMS = (int)(Time.time * 1000.0f);
                this.aggregateTimeS = 0.0f;
                this.amSpeaking = true;

                //Play the audio associated with the visemes through the head
                this.headAudioSource.PlayNow(nac);
            }
            else
            {
                if (this.debugMessagesOn.val)
                    SuperController.LogMessage("VisemeAligner: Possible Bug. Failed to find a gentle decode for \"" + audioDisplayName + "\"");
            }

        }

        //Generate a closure that will start a coroutine to speak a specific item of audio
        public Action createSpeakingDelegate(string audioDisplayName)
        {
            Action spclosure = delegate ()
            {
                if (this.debugMessagesOn.val)
                    SuperController.LogMessage("VisemeAligner: About to start speaking:" + audioDisplayName);
                StartCoroutine("startSpeaking", audioDisplayName);

            };

            return spclosure;
        }

        public void Start() {

			try {          			
				Atom containingPersonAtom = containingAtom;
				if (containingPersonAtom.type == "Person")
				{
					this.headAudioSource = containingAtom.GetStorableByID("HeadAudioSource") as AudioSourceControl;
					if(headAudioSource == null)
						SuperController.LogError("VisemeAligner: Can't get the head audio source.");
				
					//Locate the morphs on the character
					setUpMorphs(containingPersonAtom);
					//Convert the gentle JSON to an internal object that references the relevant morphs involved in the speaking.
					//This is done up front so the system has to do less work on the fly
					StartCoroutine("convertGentleJSONToAlignmentObjects");					
				}
				else
				{
					SuperController.LogError($"VisemeAligner: Plug-in must be attach to a 'Person' atom only and cannot be attached to a '{containingAtom.type}'.");
				}			
            }
			catch (Exception e) {
				SuperController.LogError("VisemeAligner: Unexpected exception caught in Start: " + e);
			}
		}

        //Silences all visemes currently in use
        void silenceCurrentlySpeakingVisemes()
        {
            this.amSpeaking = false;

            if ( this.activelySpoken != null)
            {
                if(this.activelySpoken.Count > 0){
                    foreach (Vis2FloatSeqObj vis in this.activelySpoken.Values)
                    {
                        vis.viseme.val = 0.0f;
                    }
                }
            }

            this.activelySpoken = new Hashtable();
            this.headAudioSource.Stop();
        }

        //The purpose of this is to try and make the lip transitions a little less robotic
        //by using a quadratic function to interpolate instead of linear
        //This takes adavntage of the fact that multiplying a number between 0 - 1 by itself
        //also produces a number between 0 - 1
        float quadraticLerp(float lerpIn)
        {
            return lerpIn * lerpIn;
        }


		protected void Update() {

            if (this.amSpeaking)
            {
                aggregateTimeS += Time.unscaledDeltaTime;
                int speechTime = (int)(aggregateTimeS * 1000.0f);

                //Look for visemes that should currently be being spoken based on the current real time
                //if any are found then add them to the list of current visemes to be spoken
                while (this.currentViseme < this.currentVis2FloatSeq.Count)
                {
                    if (speechTime < this.currentVis2FloatSeq[this.currentViseme].riseStartMS)
                        break;

                    if( speechTime >= this.currentVis2FloatSeq[this.currentViseme].riseStartMS &&
                        speechTime <= this.currentVis2FloatSeq[this.currentViseme].endTimeMS)
                    {
                        this.activelySpoken.Add(this.currentViseme, this.currentVis2FloatSeq[this.currentViseme]);
                        
                    }

                    this.currentViseme++;
                     
                }

                //If there are any visemes currently being spoken - animate them
                speakVisemes(speechTime);

            }
        }

        /*
        * What this is doing is on each frame, it slowly morphs the lips towards the next viseme whilst simultaneously morphing
        * away from previous viseme. So the next viseme reaches a maximum at the time when it is being spoken and the previous
        * viseme reaches a minimum. This is done quadratically so that most of the change happens late.
        */

        protected void speakVisemes(int speechTime)
        {
            List<int> keysDoneWith = new List<int>();

            //Switch on the regular VAM jaw movement when the character is not speaking
            if (this.activelySpoken.Keys.Count < 1)
            {
                this.jawAudioMultiplier.val = this.jawDriveMultiplierWhenNotSpeaking.val;
            }
            else
            {
                this.jawAudioMultiplier.val = 0.0f;
            }

            //Adjust the visemes that are currently being spoken
            foreach (int currentVisKey in this.activelySpoken.Keys)
            {
                Vis2FloatSeqObj vis = (Vis2FloatSeqObj)this.activelySpoken[currentVisKey];

                if ( speechTime <= vis.endTimeMS)
                {
                    //We are tending towards the viseme
                    if( speechTime >= vis.riseStartMS && speechTime <= vis.maximumMS)
                    {

                        float lerpDiff = (float)(speechTime - vis.riseStartMS) / (float)(vis.maximumMS - vis.riseStartMS);
                        float nextVismeLerp = Mathf.Lerp(0.0f, 1.0f, quadraticLerp(lerpDiff));
                        //float nextVismeLerp = Mathf.Lerp(0.0f, 1.0f, Mathf.SmoothStep(0.0f, 1.0f, Mathf.SmoothStep(0.0f, 1.0f, lerpDiff)));
                        if (nextVismeLerp > 0.0f && nextVismeLerp <1.0f)
                            vis.viseme.val = nextVismeLerp;
                    }
                    else
                    {
                        float lerpDiff = 1.0f;

                        if ( speechTime > vis.maximumMS && speechTime <= vis.endTimeMS)
                        {
                            if( vis.sustainEnd > 0)
                            {
                                if(speechTime <= vis.sustainEnd)
                                {
                                    vis.viseme.val = 1.0f;
                                }
                                else
                                {
                                    lerpDiff = (float)(speechTime - vis.sustainEnd) / (float)(vis.endTimeMS - vis.sustainEnd);
                                }
                            }
                            else
                            {
                                lerpDiff = (float)(speechTime - vis.maximumMS) / (float)(vis.endTimeMS - vis.maximumMS);
                                
                            }

                            float nextVismeLerp = Mathf.Lerp(1.0f, 0.0f, quadraticLerp(lerpDiff));

                            if (nextVismeLerp > 0.0f && nextVismeLerp < 1.0f)
                                vis.viseme.val = nextVismeLerp;

                        }
                    }

                }
                else
                {
                    //If we are past the time when this viseme if being spoken, then set it to zero
                    //And remove it from te list of visemes being spoken
                    vis.viseme.val = 0.0f;
                    keysDoneWith.Add(currentVisKey);
                }
            }

            foreach( int visKey in keysDoneWith)
            {
                this.activelySpoken.Remove(visKey);
            }

            

            //If there are no visemes left to be added to the list currently being spoken
            //And we have spoken all of them, then we are done
            if (this.currentViseme >= this.currentVis2FloatSeq.Count &&
                this.activelySpoken.Count < 1)
            {
                //Reset the user's original jaw audio multiplier value
                this.jawAudioMultiplier.val = this.jawDefaultMultiplierValue;
                this.amSpeaking = false;
            }
                

        }

     


    }
}